
// ZET030Dlg.cpp : implementation file
//

#include "pch.h"

#include "ZET030App.h"
#include "ZET030Dlg.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#endif

#define TIMER_REFRESH 2000
#define TIMER_DEVICE  1000
#define TIMER_SERVER  1500
#define TIMER_STREAM  100

IMPLEMENT_DYNCREATE(CZetServerCtrl, CWnd)

const char *CZET030Dlg::config_path = "conf.xml";
const char *CZET030Dlg::eth_addr = "192.168.1.100/24";
const std::vector<uint32_t> CZET030Dlg::freq_table = { 1000, 3125, 6250, 12500, 25000, 50000, 100000, 200000, 400000 };

CZET030Dlg::CZET030Dlg(CWnd* pParent /*=nullptr*/)
	: CDialogEx(IDD_ZET030_DIALOG, pParent)
{
	m_icon = AfxGetApp()->LoadIcon(IDR_MAINFRAME);
	m_icon_input_off = (HICON)LoadImage(AfxGetInstanceHandle(), MAKEINTRESOURCE(IDI_INPUT_OFF), IMAGE_ICON, 16, 16, LR_DEFAULTCOLOR);
	m_icon_input_on  = (HICON)LoadImage(AfxGetInstanceHandle(), MAKEINTRESOURCE(IDI_INPUT_ON),  IMAGE_ICON, 16, 16, LR_DEFAULTCOLOR);
	m_icon_input_icp = (HICON)LoadImage(AfxGetInstanceHandle(), MAKEINTRESOURCE(IDI_INPUT_ICP), IMAGE_ICON, 16, 16, LR_DEFAULTCOLOR);
}

CZET030Dlg::~CZET030Dlg()
{
}

BEGIN_MESSAGE_MAP(CZET030Dlg, CDialogEx)
	ON_WM_PAINT()
	ON_WM_QUERYDRAGICON()
	ON_WM_TIMER()
	ON_WM_CLOSE()
	ON_CBN_SELENDOK(IDC_COMBO_DEVICE, &CZET030Dlg::OnCbnSelendokComboDevice)
	ON_CBN_EDITUPDATE(IDC_COMBO_DEVICE, &CZET030Dlg::OnCbnEditupdateComboDevice)
	ON_BN_CLICKED(IDC_BUTTON_CONNECT, &CZET030Dlg::OnBnClickedButtonConnect)
	ON_STN_CLICKED(IDC_STATIC_ICON1 + 0, &CZET030Dlg::OnStnClickedStaticIcon1)
	ON_STN_CLICKED(IDC_STATIC_ICON1 + 1, &CZET030Dlg::OnStnClickedStaticIcon2)
	ON_STN_CLICKED(IDC_STATIC_ICON1 + 2, &CZET030Dlg::OnStnClickedStaticIcon3)
	ON_STN_CLICKED(IDC_STATIC_ICON1 + 3, &CZET030Dlg::OnStnClickedStaticIcon4)
	ON_STN_CLICKED(IDC_STATIC_NAME1 + 0, &CZET030Dlg::OnStnClickedStaticIcon1)
	ON_STN_CLICKED(IDC_STATIC_NAME1 + 1, &CZET030Dlg::OnStnClickedStaticIcon2)
	ON_STN_CLICKED(IDC_STATIC_NAME1 + 2, &CZET030Dlg::OnStnClickedStaticIcon3)
	ON_STN_CLICKED(IDC_STATIC_NAME1 + 3, &CZET030Dlg::OnStnClickedStaticIcon4)
	ON_BN_CLICKED(IDC_CHECK_ICP1, &CZET030Dlg::OnBnClickedCheckIcp)
	ON_BN_CLICKED(IDC_CHECK_ICP2, &CZET030Dlg::OnBnClickedCheckIcp)
	ON_BN_CLICKED(IDC_CHECK_ICP3, &CZET030Dlg::OnBnClickedCheckIcp)
	ON_BN_CLICKED(IDC_CHECK_ICP4, &CZET030Dlg::OnBnClickedCheckIcp)
	ON_BN_CLICKED(IDC_CHECK_AMP1, &CZET030Dlg::OnBnClickedCheckAmp)
	ON_BN_CLICKED(IDC_CHECK_AMP2, &CZET030Dlg::OnBnClickedCheckAmp)
	ON_BN_CLICKED(IDC_CHECK_AMP3, &CZET030Dlg::OnBnClickedCheckAmp)
	ON_BN_CLICKED(IDC_CHECK_AMP4, &CZET030Dlg::OnBnClickedCheckAmp)
	ON_EN_CHANGE(IDC_EDIT_SENSE1, &CZET030Dlg::OnEnChangeEditSense)
	ON_EN_CHANGE(IDC_EDIT_SENSE2, &CZET030Dlg::OnEnChangeEditSense)
	ON_EN_CHANGE(IDC_EDIT_SENSE3, &CZET030Dlg::OnEnChangeEditSense)
	ON_EN_CHANGE(IDC_EDIT_SENSE4, &CZET030Dlg::OnEnChangeEditSense)
	ON_EN_CHANGE(IDC_EDIT_UNIT1, &CZET030Dlg::OnEnChangeEditUnit)
	ON_EN_CHANGE(IDC_EDIT_UNIT2, &CZET030Dlg::OnEnChangeEditUnit)
	ON_EN_CHANGE(IDC_EDIT_UNIT3, &CZET030Dlg::OnEnChangeEditUnit)
	ON_EN_CHANGE(IDC_EDIT_UNIT4, &CZET030Dlg::OnEnChangeEditUnit)
	ON_CBN_SELCHANGE(IDC_COMBO_FREQ, &CZET030Dlg::OnCbnSelchangeComboFreq)
	ON_CBN_SELCHANGE(IDC_COMBO_TEST, &CZET030Dlg::OnCbnSelchangeComboTest)
	ON_CBN_SELCHANGE(IDC_COMBO_ETH_METHOD, &CZET030Dlg::OnCbnSelchangeComboEthMethod)
	ON_EN_CHANGE(IDC_EDIT_ETH_ADDR, &CZET030Dlg::OnEnChangeEditEthAddr)
	ON_BN_CLICKED(IDC_CHECK_ETH_FTP, &CZET030Dlg::OnBnClickedCheckEthFtp)
	ON_CBN_SELCHANGE(IDC_COMBO_REC_MODE, &CZET030Dlg::OnCbnSelchangeComboRecMode)
	ON_EN_CHANGE(IDC_EDIT_REC_MINUTES, &CZET030Dlg::OnEnChangeEditRecMinutes)
	ON_BN_CLICKED(IDC_BUTTON_APPLY, &CZET030Dlg::OnBnClickedButtonApply)
	ON_BN_CLICKED(IDC_BUTTON_REVERT, &CZET030Dlg::OnBnClickedButtonRevert)
END_MESSAGE_MAP()

void CZET030Dlg::DoDataExchange(CDataExchange *pDX)
{
	int i;

	CDialogEx::DoDataExchange(pDX);

	DDX_Control(pDX, IDC_COMBO_DEVICE, m_ui.device);
	DDX_Control(pDX, IDC_BUTTON_CONNECT, m_ui.connect);
	DDX_Control(pDX, IDC_STATUS, m_ui.status);

	for (i = 0; i < channel_quantity; i++)
	{
		DDX_Control(pDX, IDC_STATIC_ICON1 + i, m_ui.channel[i].icon);
		DDX_Control(pDX, IDC_STATIC_NAME1 + i, m_ui.channel[i].name);
		DDX_Control(pDX, IDC_STATIC_LEVEL1 + i, m_ui.channel[i].level);
		DDX_Control(pDX, IDC_CHECK_AMP1 + i, m_ui.channel[i].amp);
		DDX_Control(pDX, IDC_CHECK_ICP1 + i, m_ui.channel[i].icp);
		DDX_Control(pDX, IDC_EDIT_SENSE1 + i, m_ui.channel[i].sense);
		DDX_Control(pDX, IDC_EDIT_UNIT1 + i, m_ui.channel[i].unit);
	}

	DDX_Control(pDX, IDC_STATIC_FREQ, m_ui.freq_label);
	DDX_Control(pDX, IDC_COMBO_FREQ, m_ui.freq);
	DDX_Control(pDX, IDC_STATIC_TEST, m_ui.test_label);
	DDX_Control(pDX, IDC_COMBO_TEST, m_ui.test);

	DDX_Control(pDX, IDC_STATIC_ETH_ADDR, m_ui.eth.label);
	DDX_Control(pDX, IDC_COMBO_ETH_METHOD, m_ui.eth.method);
	DDX_Control(pDX, IDC_EDIT_ETH_ADDR, m_ui.eth.addr);
	DDX_Control(pDX, IDC_CHECK_ETH_FTP, m_ui.eth.ftp);

	DDX_Control(pDX, IDC_STATIC_REC_MODE, m_ui.rec_mode_label);
	DDX_Control(pDX, IDC_COMBO_REC_MODE, m_ui.rec_mode);
	DDX_Control(pDX, IDC_STATIC_REC_MINUTES, m_ui.rec_minutes_label);
	DDX_Control(pDX, IDC_EDIT_REC_MINUTES, m_ui.rec_minutes);

	DDX_Control(pDX, IDC_BUTTON_APPLY, m_ui.apply);
	DDX_Control(pDX, IDC_BUTTON_REVERT, m_ui.revert);
}

BOOL CZET030Dlg::OnInitDialog()
{
	CDialogEx::OnInitDialog();

	// Set the icon for this dialog.  The framework does this automatically
	//  when the application's main window is not a dialog
	SetIcon(m_icon, TRUE);		// Set big icon
	SetIcon(m_icon, FALSE);		// Set small icon

	CString s;

	m_ui.freq.ResetContent();
	for (auto freq : freq_table)
	{
		CString unit;

		if ((freq < 1000) || (freq % 1000) != 0)
		{
			(void)unit.LoadString(IDS_HZ);
		} else
		{
			freq /= 1000;
			(void)unit.LoadString(IDS_KHZ);
		}

		CString s;
		s.Format(L"%u %s", freq, unit.GetString());
		m_ui.freq.AddString(s.GetString());
	}
	m_ui.freq.SetCurSel(0);

	m_ui.test.ResetContent();
	(void)s.LoadString(IDS_TEST_OFF);
	m_ui.test.AddString(s.GetString());
	(void)s.LoadString(IDS_TEST_SQR);
	m_ui.test.AddString(s.GetString());
	(void)s.LoadString(IDS_TEST_SHORT);
	m_ui.test.AddString(s.GetString());
	(void)s.LoadString(IDS_TEST_NEG);
	m_ui.test.AddString(s.GetString());
	m_ui.test.SetCurSel(0);

	m_ui.eth.method.ResetContent();
	(void)s.LoadString(IDS_ETH_DHCP);
	m_ui.eth.method.AddString(s.GetString());
	(void)s.LoadString(IDS_ETH_STATIC);
	m_ui.eth.method.AddString(s.GetString());
	m_ui.eth.method.SetCurSel(0);

	m_ui.rec_mode.ResetContent();
	(void)s.LoadString(IDS_REC_BUTTON);
	m_ui.rec_mode.AddString(s.GetString());
	(void)s.LoadString(IDS_REC_AUTO);
	m_ui.rec_mode.AddString(s.GetString());
	m_ui.rec_mode.SetCurSel(0);

	RefreshDeviceList();
	ReportStatus(IDS_DEVICE_NOT_SELECTED);
	UpdateControls();

	SetTimer(TIMER_REFRESH, TIMER_REFRESH, NULL);
	SetTimer(TIMER_SERVER, TIMER_SERVER, NULL);

	return TRUE;
}

BOOL CZET030Dlg::DestroyWindow()
{
	CloseServer();
	CloseDevice();

	return CDialogEx::DestroyWindow();
}

// The system calls this function to obtain the cursor to display while the user drags
//  the minimized window.
HCURSOR CZET030Dlg::OnQueryDragIcon()
{
	return static_cast<HCURSOR>(m_icon);
}

// If you add a minimize button to your dialog, you will need the code below
//  to draw the icon.  For MFC applications using the document/view model,
//  this is automatically done for you by the framework.

void CZET030Dlg::OnPaint()
{
	if (IsIconic())
	{
		CPaintDC dc(this); // device context for painting

		SendMessage(WM_ICONERASEBKGND, reinterpret_cast<WPARAM>(dc.GetSafeHdc()), 0);

		// Center icon in client rectangle
		int cxIcon = GetSystemMetrics(SM_CXICON);
		int cyIcon = GetSystemMetrics(SM_CYICON);
		CRect rect;
		GetClientRect(&rect);
		int x = (rect.Width() - cxIcon + 1) / 2;
		int y = (rect.Height() - cyIcon + 1) / 2;

		// Draw the icon
		dc.DrawIcon(x, y, m_icon);
	} else
	{
		CDialogEx::OnPaint();
	}
}

void CZET030Dlg::OnOK()
{
}

void CZET030Dlg::OnCancel()
{
}

void CZET030Dlg::OnClose()
{
	// Bypass CDialogEx::OnClose calling OnCancel
	EndDialog(0);
}

void CZET030Dlg::OnTimer(UINT_PTR nIDEvent)
{
	switch (nIDEvent)
	{
	case TIMER_REFRESH:
		RefreshDeviceList();
		break;
	case TIMER_DEVICE:
		WatchDevice();
		break;
	case TIMER_SERVER:
		WatchServer();
		break;
	case TIMER_STREAM:
		WatchStream();
		break;
	}

	CDialogEx::OnTimer(nIDEvent);
}

void CZET030Dlg::OnCbnSelendokComboDevice()
{
	int pos = m_ui.device.GetCurSel();
	if (pos >= 0)
	{
		CString s;
		m_ui.device.GetLBText(pos, s);
		m_ui.device.SetWindowText(s.GetString());
	}

	m_ui.device.ShowDropDown(FALSE);
	UpdateControls();
}

void CZET030Dlg::OnCbnEditupdateComboDevice()
{
	UpdateControls();
}

void CZET030Dlg::OnBnClickedButtonConnect()
{
	if (m_device_path_selected.empty())
	{
		m_device_path_selected = GetDevicePath();
		if (!m_device_path_selected.empty())
		{
			CString s;
			(void)s.LoadString(IDS_DISCONNECT);
			m_ui.connect.SetWindowText(s.GetString());
			OpenDevice();
			SetTimer(TIMER_DEVICE, TIMER_DEVICE, NULL);
		}
	}
	else
	{
		CloseDevice();
		KillTimer(TIMER_DEVICE);
		m_device_path_selected.clear();

		CString s;
		(void)s.LoadString(IDS_CONNECT);
		m_ui.connect.SetWindowText(s.GetString());

		SetWindowText(L"ZET 030");

		RefreshDeviceList();
	}

	UpdateControls();
}

void CZET030Dlg::OnStnClickedStaticIcon1()
{
	ToggleChannelControl(0);
}

void CZET030Dlg::OnStnClickedStaticIcon2()
{
	ToggleChannelControl(1);
}

void CZET030Dlg::OnStnClickedStaticIcon3()
{
	ToggleChannelControl(2);
}

void CZET030Dlg::OnStnClickedStaticIcon4()
{
	ToggleChannelControl(3);
}

void CZET030Dlg::OnBnClickedCheckIcp()
{
	for (size_t i = 0; i < channel_quantity; i++)
	{
		m_cfg_temp.channel[i].icp = (m_ui.channel[i].icp.GetCheck() == BST_CHECKED);
	}
	UpdateControls(true);
}

void CZET030Dlg::OnBnClickedCheckAmp()
{
	for (size_t i = 0; i < channel_quantity; i++)
	{
		m_cfg_temp.channel[i].amp_ratio = (m_ui.channel[i].amp.GetCheck() == BST_CHECKED) ? 30.0f : 1.0f;
	}
	UpdateControls(true);
}

void CZET030Dlg::OnEnChangeEditSense()
{
	for (size_t i = 0; i < channel_quantity; i++)
	{
		CString s;
		m_ui.channel[i].sense.GetWindowText(s);

		float sense = wcstof(s.GetString(), nullptr);
		if (sense != 0.0 && s != m_cfg_temp.channel[i].sense.c_str())
		{
			m_cfg_temp.channel[i].sense = s.GetString();
		}
	}
	UpdateControls(true);
}

void CZET030Dlg::OnEnChangeEditUnit()
{
	for (size_t i = 0; i < channel_quantity; i++)
	{
		CString s;
		m_ui.channel[i].unit.GetWindowText(s);
		s.Trim();

		if (!s.IsEmpty() && s != m_cfg_temp.channel[i].unit.c_str())
		{
			m_cfg_temp.channel[i].unit = s.GetString();
		}
	}
	UpdateControls(true);
}

void CZET030Dlg::OnCbnSelchangeComboFreq()
{
	int freq_sel = m_ui.freq.GetCurSel();
	if (freq_sel < (int)freq_table.size())
		m_cfg_temp.freq = freq_table[freq_sel];
	else
		m_cfg_temp.freq = 25000;
	UpdateControls(true);
}

void CZET030Dlg::OnCbnSelchangeComboTest()
{
	SendTest(static_cast<enum ui::test>(m_ui.test.GetCurSel()));
}

void CZET030Dlg::OnCbnSelchangeComboEthMethod()
{
	if (m_cfg_temp.eth.has_value())
	{
		m_cfg_temp.eth->dhcp = (m_ui.eth.method.GetCurSel() == 0);
		UpdateControls(true);
	}
}

void CZET030Dlg::OnEnChangeEditEthAddr()
{
	if (m_cfg_temp.eth.has_value())
	{
		CString s;
		m_ui.eth.addr.GetWindowText(s);

		m_cfg_temp.eth->addr = pugi::as_utf8(s.GetString());
		UpdateControls(true);
	}
}

void CZET030Dlg::OnBnClickedCheckEthFtp()
{
	if (m_cfg_temp.eth.has_value())
	{
		m_cfg_temp.eth->ftp = (m_ui.eth.ftp.GetCheck() == BST_CHECKED);
		UpdateControls(true);
	}
}

void CZET030Dlg::OnCbnSelchangeComboRecMode()
{
	if (m_cfg_temp.rec_button.has_value())
	{
		int pos = m_ui.rec_mode.GetCurSel();

		if (pos == 0)
			m_cfg_temp.rec_button.emplace(true);
		else if (pos == 1)
			m_cfg_temp.rec_button.emplace(false);

		UpdateControls(true);
	}
}

void CZET030Dlg::OnEnChangeEditRecMinutes()
{
	CString s;
	m_ui.rec_minutes.GetWindowText(s);

	m_cfg_temp.rec_minutes = wcstoul(s.GetString(), NULL, 0);
	UpdateControls(true);
}

void CZET030Dlg::OnBnClickedButtonApply()
{
	if (m_xml && m_device_ready && !(m_cfg == m_cfg_temp))
	{
		StopStream();
		SaveConfig();
		LoadConfig();
		StartStream();
	}
}

void CZET030Dlg::OnBnClickedButtonRevert()
{
	m_cfg_temp = m_cfg;
	UpdateControls();
}

void CZET030Dlg::ReportStatus(const CString &text)
{
	m_ui.status.SetWindowText(text);
}

void CZET030Dlg::ReportStatus(UINT ids)
{
	CString text;
	(void)text.LoadString(ids);
	m_ui.status.SetWindowText(text);
}

void CZET030Dlg::UpdateControls(bool skip_set_value)
{
	if (m_device_path_selected.empty())
	{
		m_ui.device.EnableWindow(TRUE);
		if (!GetDevicePath().empty())
		{
			ReportStatus(IDS_DEVICE_SELECTED);
			m_ui.connect.EnableWindow(TRUE);
		}
		else
		{
			ReportStatus(IDS_DEVICE_NOT_SELECTED);
			m_ui.connect.EnableWindow(FALSE);
		}
	}
	else
	{
		m_ui.device.EnableWindow(FALSE);
	}

	BOOL ready = m_device_ready;

	for (size_t i = 0; i < channel_quantity; i++)
	{
		auto &ui = m_ui.channel[i];

		ui.icon.EnableWindow(ready);
		ui.name.EnableWindow(ready);
		ui.level.EnableWindow(ready);
		ui.icp.EnableWindow(ready);
		ui.amp.EnableWindow(ready);
		ui.sense.EnableWindow(ready);
		ui.unit.EnableWindow(ready);
	}

	m_ui.freq_label.EnableWindow(ready);
	m_ui.freq.EnableWindow(ready);

	m_ui.test_label.EnableWindow(ready);
	m_ui.test.EnableWindow(ready);

	BOOL eth_ready = m_device_ready && m_cfg_temp.eth.has_value();

	m_ui.eth.label.EnableWindow(eth_ready);
	m_ui.eth.method.EnableWindow(eth_ready);
	m_ui.eth.addr.EnableWindow(eth_ready);
	m_ui.eth.ftp.EnableWindow(eth_ready);

	m_ui.rec_mode_label.EnableWindow(ready);
	if (m_device_ready && m_cfg_temp.rec_button.has_value())
	{
		m_ui.rec_mode.EnableWindow(TRUE);
	}
	else
	{
		m_ui.rec_mode.EnableWindow(FALSE);
		m_ui.rec_mode.SetCurSel(0);
	}

	m_ui.rec_minutes_label.EnableWindow(ready);
	m_ui.rec_minutes.EnableWindow(ready);

	if (m_device_ready)
	{
		for (size_t i = 0; i < channel_quantity; i++)
		{
			auto &cfg_temp = m_cfg_temp.channel[i];
			auto &ui = m_ui.channel[i];

			if (cfg_temp.enabled)
			{
				ui.icon.SetIcon(cfg_temp.icp ? m_icon_input_icp : m_icon_input_on);
			}
			else
			{
				ui.icon.SetIcon(m_icon_input_off);
				ui.level.SetWindowText(L"~");
			}

			if (!skip_set_value)
			{
				ui.icp.SetCheck(cfg_temp.icp ? BST_CHECKED : BST_UNCHECKED);
				ui.amp.SetCheck(cfg_temp.amp_ratio > 1.0f ? BST_CHECKED : BST_UNCHECKED);
				ui.sense.SetWindowText(cfg_temp.sense.c_str());
				ui.unit.SetWindowText(cfg_temp.unit.c_str());
			}
		}

		if (!skip_set_value)
		{
			for (size_t i = 0; i < freq_table.size(); i++)
			{
				if (freq_table[i] == m_cfg_temp.freq)
				{
					m_ui.freq.SetCurSel((int)i);
					break;
				}
			}

			if (m_cfg_temp.eth.has_value())
			{
				m_ui.eth.method.SetCurSel(m_cfg_temp.eth->dhcp ? 0 : 1);
				if (!m_cfg_temp.eth->addr.empty())
					m_ui.eth.addr.SetWindowText(pugi::as_wide(m_cfg_temp.eth->addr).c_str());
				m_ui.eth.ftp.SetCheck(m_cfg_temp.eth->ftp ? BST_CHECKED : BST_UNCHECKED);
			}

			if (m_cfg_temp.rec_button.has_value())
			{
				if (*m_cfg_temp.rec_button)
					m_ui.rec_mode.SetCurSel(0);
				else
					m_ui.rec_mode.SetCurSel(1);
			}

			CString s;
			s.Format(L"%u", m_cfg_temp.rec_minutes);
			m_ui.rec_minutes.SetWindowText(s.GetString());
		}
	}
	else
	{
		for (size_t i = 0; i < channel_quantity; i++)
		{
			m_ui.channel[i].icon.SetIcon(m_icon_input_off);
		}
	}

	BOOL changed = m_device_ready && !(m_cfg == m_cfg_temp);
	m_ui.apply.EnableWindow(changed);
	m_ui.revert.EnableWindow(changed);
}

void CZET030Dlg::ToggleChannelControl(size_t ch)
{
	if (ch < channel_quantity)
	{
		m_cfg_temp.channel[ch].enabled = !m_cfg_temp.channel[ch].enabled;

		size_t enabled_count = 0;
		for (size_t i = 0; i < channel_quantity; i++)
		{
			if (m_cfg_temp.channel[i].enabled)
				enabled_count++;
		}
		if (enabled_count == 0)
			m_cfg_temp.channel[ch].enabled = true;
		UpdateControls();
	}
}

bool CZET030Dlg::LoadConfig(void)
{
	if (!m_device_ready)
		return false;

	auto load_cb = [](struct zet030_device *d, void *arg, uint32_t offset, const void *data, int size) -> int
	{
		auto text = static_cast<std::vector<uint8_t> *>(arg);
		auto data8 = static_cast<const uint8_t *>(data);

		text->insert(text->end(), data8, data8 + size);
		return 0;
	};

	std::vector<uint8_t> xmltext;
	int r = zet030_load(m_device, config_path, load_cb, &xmltext);
	if (r < 0)
	{
		ReportStatus(IDS_CONFIG_LOAD_FAILED);
		return false;
	}

	auto result = m_xml.load_buffer(xmltext.data(), xmltext.size());
	if (result.status != pugi::xml_parse_status::status_ok)
	{
		ReportStatus(IDS_CONFIG_WRONG_FORMAT);
		return false;
	}

	pugi::xpath_node xn;

	xn = m_xml.select_node("/Config/Device/@type");
	if (!xn || xn.attribute().as_int() != device_type)
	{
		ReportStatus(IDS_CONFIG_WRONG_TYPE);
		return false;
	}

	m_device_name = L"ZET 030";

	xn = m_xml.select_node("/Config/Device/@name");
	if (xn)
	{
		m_device_name = pugi::as_wide(xn.attribute().as_string()).c_str();
	}

	xn = m_xml.select_node("/Config/Device/@serial");
	if (xn && xn.attribute().as_uint() > 0)
	{
		m_device_name.AppendFormat(L" %u", xn.attribute().as_uint());

		SetWindowText(m_device_name.GetString());
	}
	else
	{
		ReportStatus(IDS_CONFIG_WRONG_SERIAL);
		return false;
	}

	m_cfg.eth.reset();
	if (m_xml.select_node("/Config/Device/Ethernet"))
	{
		m_cfg.eth.emplace();

		xn = m_xml.select_node("/Config/Device/Ethernet/@method");
		if (xn)
		{
			std::string method(xn.attribute().as_string());
			m_cfg.eth->dhcp = (method == "auto" || method == "dhcp");
		}

		xn = m_xml.select_node("/Config/Device/Ethernet/@addr");
		if (xn)
		{
			m_cfg.eth->addr = xn.attribute().as_string();
		}

		xn = m_xml.select_node("/Config/Device/Ethernet/@ftp");
		if (xn)
		{
			std::string ftp(xn.attribute().as_string());
			m_cfg.eth->ftp = (ftp == "yes" || ftp == "on");
		}
	}

	xn = m_xml.select_node("/Config/Device/Freq");
	m_cfg.freq = xn.node().text().as_int(25000);

	xn = m_xml.select_node("/Config/Device/Channel");
	if (xn)
	{
		int channel_mask = xn.node().text().as_int();
		for (size_t i = 0; i < channel_quantity; i++)
		{
			bool enabled = channel_mask & (1u << i);
			m_cfg.channel[i].enabled = enabled;
		}
	}

	xn = m_xml.select_node("/Config/Device/HCPChannel");
	if (xn)
	{
		int hcp_channel_mask = xn.node().text().as_int();
		for (size_t i = 0; i < channel_quantity; i++)
		{
			bool icp = hcp_channel_mask & (1u << i);
			m_cfg.channel[i].icp = icp;
		}
	}

	xn = m_xml.select_node("/Config/Device/KodAmplify");;
	if (xn)
	{
		size_t count = 0;
		std::string kodampl_str = xn.node().text().as_string();
		const char *str = kodampl_str.data();
		while (count < channel_quantity)
		{
			if (count != 0)
			{
				while (*str == ' ')
					str++;
				if (*str != ',')
				{
					count = 0;
					break;
				}
				str++;
			}
			char *end;
			uint32_t kod = strtoul(str, &end, 0);
			str = end;
			m_cfg.channel[count++].amp_ratio = kod > 0 ? 30.0f : 1.0f;
		}
	}

	xn = m_xml.select_node("/Config/Device/DigitalResolChanADC");;
	if (xn)
	{
		size_t count = 0;
		std::string resol_str = xn.node().text().as_string();
		const char *str = resol_str.data();
		while (count < channel_quantity)
		{
			if (count != 0)
			{
				while (*str == ' ')
					str++;
				if (*str != ',')
				{
					count = 0;
					break;
				}
				str++;
			}
			char *end;
			m_cfg.channel[count++].lsb = strtof(str, &end); // volts
			str = end;
		}
	}

	for (size_t i = 0; i < channel_quantity; i++)
	{
		auto &cfg = m_cfg.channel[i];

		CString s;
		s.Format(L"/Config/Device/Channels/Channel[@id='%zd']/@units", i);
		xn = m_xml.select_node(pugi::as_utf8(s.GetString()).c_str());
		if (xn)
			cfg.unit = pugi::as_wide(xn.attribute().as_string());

		s.Format(L"/Config/Device/Channels/Channel[@id='%zd']/Sense", i);
		xn = m_xml.select_node(pugi::as_utf8(s.GetString()).c_str());
		if (xn)
			cfg.sense = pugi::as_wide(xn.node().text().as_string());
	}

	m_cfg.rec_button.reset();
	m_cfg.rec_button.emplace(true);
	xn = m_xml.select_node("/Config/Device/Recorder/@start");
	if (xn)
	{
		std::string mode(xn.attribute().as_string());
		if (mode == "auto")
			m_cfg.rec_button.emplace(false);
		else if (mode == "button")
			m_cfg.rec_button.emplace(true);
	}

	m_cfg.rec_minutes = 10;
	xn = m_xml.select_node("/Config/Device/RecordMinutes");
	if (xn)
	{
		m_cfg.rec_minutes = xn.node().text().as_int();
	}

	m_state.freq = (float)m_cfg.freq;
	m_state.channel_count = 0;
	for (size_t i = 0; i < channel_quantity; i++)
	{
		auto &cfg = m_cfg.channel[i];

		if (cfg.enabled)
			m_state.channel_count++;

		m_state.channel[i].enabled = cfg.enabled;

		float sense = wcstof(cfg.sense.c_str(), nullptr);
		if (sense == 0)
			sense = 1.0f;
		m_state.channel[i].int_to_float = cfg.lsb / cfg.amp_ratio / sense;

		// round max_level to two most significant decimal digits, e.g.:
		//  10123.456 -> 10000.0
		//   9998.123 -> 10000.0
		//   9090     ->  9100.0
		//   9.123    ->     9.1
		//   0.90912  ->     0.9
		float max_volt = fabsf(m_state.channel[i].int_to_float * powf(2, 31));
		float fix = powf(10.0f, log10(max_volt) - 1);
		m_state.channel[i].max_level = round(max_volt / fix) * fix;
	}

	m_cfg_temp = m_cfg;
	UpdateControls();
	UpdateServerChannels();
	return true;
}

bool CZET030Dlg::SaveConfig()
{
	if (m_cfg_temp == m_cfg)
		return true; // no changes

	auto ensure_node = [](pugi::xml_node node, const char *name) -> pugi::xml_node
	{
		auto child = node.child(name);
		if (!child)
			child = node.append_child(name);
		return child;
	};

	auto ensure_attr = [](pugi::xml_node node, const char *name) -> pugi::xml_attribute
	{
		auto attr = node.attribute(name);
		if (!attr)
			attr = node.append_attribute(name);
		return attr;
	};

	CStringA s;
	pugi::xml_node node;
	pugi::xml_attribute attr;
	uint32_t mask;

	auto device_node = m_xml.first_child().child("Device");

	if (m_cfg_temp.eth.has_value())
	{
		node = ensure_node(device_node, "Ethernet");

		attr = ensure_attr(node, "method");
		attr.set_value(m_cfg_temp.eth->dhcp ? "dhcp" : "static");

		attr = ensure_attr(node, "addr");
		if (!m_cfg_temp.eth->addr.empty())
			attr.set_value(m_cfg_temp.eth->addr.c_str());

		attr = ensure_attr(node, "ftp");
		attr.set_value(m_cfg_temp.eth->ftp ? "yes" : "no");
	}

	node = ensure_node(device_node, "Freq");
	node.text().set(m_cfg_temp.freq);

	mask = 0;
	for (size_t i = 0; i < channel_quantity; i++)
	{
		if (m_cfg_temp.channel[i].enabled)
			mask |= (1u << i);
	}
	s.Format("0x%04x", mask);
	node = ensure_node(device_node, "Channel");
	node.text().set(s.GetString());

	mask = 0;
	for (size_t i = 0; i < channel_quantity; i++)
	{
		if (m_cfg_temp.channel[i].icp)
			mask |= (1u << i);
	}
	s.Format("0x%04x", mask);
	node = ensure_node(device_node, "HCPChannel");
	node.text().set(s.GetString());

	s.Format("%u", m_cfg_temp.channel[0].amp_ratio > 1.0f ? 1 : 0);
	for (size_t i = 1; i < channel_quantity; i++)
	{
		s.AppendFormat(",%u", m_cfg_temp.channel[i].amp_ratio > 1.0f ? 1 : 0);
	}
	node = ensure_node(device_node, "KodAmplify");
	node.text().set(s.GetString());

	if (m_cfg_temp.rec_button.has_value())
	{
		node = ensure_node(device_node, "Recorder");
		attr = ensure_attr(node, "start");
		attr.set_value(*m_cfg_temp.rec_button ? "button" : "auto");
	}

	node = ensure_node(device_node, "RecordMinutes");
	node.text().set(m_cfg_temp.rec_minutes);

	auto channels_node = ensure_node(device_node, "Channels");

	auto ensure_channel_node = [this,&channels_node](int id) -> pugi::xml_node
	{
		for (auto node = channels_node.child("Channel"); node; node = node.next_sibling("Channel"))
		{
			auto id_attr = node.attribute("id");
			if (id_attr.as_int(-1) == id)
				return node;
		}

		CString s;
		s.Format(L"%s #%d", m_device_name.GetString(), id + 1);

		auto node = channels_node.append_child("Channel");
		node.append_attribute("id").set_value(id);
		node.append_attribute("name").set_value(s.GetString());
		return node;
	};

	for (int i = 0; i < channel_quantity; i++)
	{
		auto channel_node = ensure_channel_node(i);
		auto &cfg_temp = m_cfg_temp.channel[i];

		auto attr = ensure_attr(channel_node, "units");
		attr.set_value(pugi::as_utf8(cfg_temp.unit).c_str());

		node = ensure_node(channel_node, "Sense");
		node.text().set(pugi::as_utf8(cfg_temp.sense).c_str());
	}

	std::ostringstream os;
	m_xml.save(os, "  ");

	auto save_cb = [](struct zet030_device *d, void *arg, uint32_t offset, void *buf, int max_size) -> int
	{
		auto text = static_cast<std::string *>(arg);

		if (offset >= text->size())
			return 0;

		int size = static_cast<int>(text->size() - offset);
		if (size > max_size)
			size = max_size;
		memcpy(buf, text->data() + offset, size);
		return size;
	};

	std::string xmltext = os.str();
	int r = zet030_save(m_device, config_path, save_cb, &xmltext);
	return true;
}

bool CZET030Dlg::SendTest(enum ui::test test)
{
	if (!m_device_ready)
		return false;

	char response[256];
	int r = -1;

	switch (test)
	{
	case ui::test_off:
		r = zet030_talk(m_device, "test off", response, sizeof(response));
		break;
	case ui::test_sqr:
		r = zet030_talk(m_device, "test sqr", response, sizeof(response));
		break;
	case ui::test_short:
		r = zet030_talk(m_device, "test short", response, sizeof(response));
		break;
	case ui::test_neg:
		r = zet030_talk(m_device, "test neg", response, sizeof(response));
		break;
	case ui::test_sqr_short:
		r = zet030_talk(m_device, "test sqr+short", response, sizeof(response));
		break;
	}

	return r == 0;
}

bool CZET030Dlg::StartStream()
{
	if (m_state.started)
		return true;

	auto stream_cb = [](struct zet030_device *d, void *arg, const int32_t *data, int count, const struct zet030_stream_time *time) -> int
	{
		CZET030Dlg *self = static_cast<CZET030Dlg *>(arg);

		self->ReceiveStream(data, count);
		return 0;
	};

	m_state.mutex.lock();
	m_state.fifo.clear();
	m_state.mutex.unlock();

	m_state.started = true;
	if (zet030_start(m_device, stream_cb, this) < 0)
	{
		m_state.started = false;
		return false;
	}

	SetTimer(TIMER_STREAM, TIMER_STREAM, NULL);
	return true;
}

bool CZET030Dlg::StopStream()
{
	if (m_state.started)
	{
		KillTimer(TIMER_STREAM);

		m_state.started = false;
		zet030_stop(m_device);
	}
	return false;
}

void CZET030Dlg::ReceiveStream(const int32_t *data, int count)
{
	if (m_state.started)
	{
		m_state.mutex.lock();
		m_state.fifo.insert(m_state.fifo.end(), data, data + count);
		m_state.mutex.unlock();
	}
}

void CZET030Dlg::WatchStream()
{
	std::vector<int32_t> fifo;

	m_state.mutex.lock();
	fifo.swap(m_state.fifo);
	m_state.mutex.unlock();

	if (!m_device_ready)
		return;

	size_t fifo_size = fifo.size();
	if (fifo_size == 0)
		return;

	long frame_count = static_cast<long>(fifo_size / m_state.channel_count);
	std::vector<float> data(frame_count);

	m_state.counter64 += frame_count;
#if 1
	double time = static_cast<double>(m_state.counter64);
	time = m_state.time + time / m_state.freq;
#else
	m_state.time += frame_count / m_state.freq;
	double time = m_state.time;
#endif

	const int32_t *ch_fifo = fifo.data();

	for (int ch = 0; ch < channel_quantity; ch++)
	{
		auto &state = m_state.channel[ch];
		if (state.enabled)
		{
			float ratio = state.int_to_float;

			for (long i = 0; i < frame_count; i++)
			{
				float f = ch_fifo[i * m_state.channel_count] * ratio;

				data[i] = f;

				state.mean_acc += f;
				state.rms_acc += (f * f);
				if (++state.stat_cnt >= m_cfg.freq)
				{
					float level = static_cast<float>(sqrt(state.rms_acc / state.stat_cnt));
					state.mean_acc = 0;
					state.rms_acc = 0;
					state.stat_cnt = 0;

					CString s;
					s.Format(L"%+0.3f", level);
					m_ui.channel[ch].level.SetWindowText(s);
				}
			}

			if (m_srv)
			{
				if (m_srv->PutData(ch, time, frame_count, data.data()) != 0)
					CloseServer();
			}

			ch_fifo++; // next channel
		}
	}
}

void CZET030Dlg::RefreshDeviceList()
{
	if (!m_device_path_selected.empty())
		return;
	if (m_ui.device.GetDroppedState())
		return;

	auto scan_cb = [](void *arg, const char *path, uint16_t pid, uint32_t serial) -> int
	{
		std::list<CString> *device_list = static_cast<std::list<CString> *>(arg);

		CString s;

		if (serial != 0)
			s.Format(L"%hs (%u)", path, serial);
		else
			s.Format(L"%hs (busy)", path);

		device_list->push_back(s);
		return 0;
	};

	std::list<CString> device_list;
	zet030_scan(scan_cb, &device_list);
	device_list.push_back(L"192.168.1.100");

	if (device_list != m_device_list)
	{
		m_device_list = device_list;

		CString cur_path;
		m_ui.device.GetWindowText(cur_path);

		m_ui.device.ResetContent();
		for (auto const &s : m_device_list)
			m_ui.device.AddString(s.GetString());

		if (!cur_path.IsEmpty())
			m_ui.device.SetWindowText(cur_path.GetString());

		UpdateControls();
	}
}

std::string CZET030Dlg::GetDevicePath()
{
	CString s;
	m_ui.device.GetWindowText(s);

	int space = s.Find(L' ');
	if (space >= 0)
		s = s.Left(space);

	return pugi::as_utf8(s.GetString());
}

bool CZET030Dlg::OpenDevice()
{
	if (m_device)
		return true;

	if (m_device_delay_counter > 0)
	{
		m_device_delay_counter--;
		return false;
	}

	if (m_device_path_selected.empty())
		return false;

	m_device_delay_counter = 5;
	m_device_ready = false;

	auto connect_cb = [](struct zet030_device *d, void *arg, int status) -> int
	{
		CZET030Dlg *self = static_cast<CZET030Dlg *>(arg);
		self->PostMessage(WM_TIMER, TIMER_DEVICE, 0);
		return 0;
	};

	m_device = zet030_connect(m_device_path_selected.c_str(), connect_cb, this);
	if (m_device)
	{
		ReportStatus(IDS_DEVICE_CONNECTING);
		return true;
	}

	m_device = zet030_open_path(m_device_path_selected.c_str());
	if (m_device)
	{
		m_device_ready = true;
		return InitDevice();
	}

	ReportStatus(IDS_DEVICE_OPEN_FAILED);
	return false;
}

bool CZET030Dlg::InitDevice()
{
	if (!m_device_ready)
		return false;

	if (!LoadConfig())
	{
		zet030_close(m_device);
		m_device = NULL;
		m_device_ready = false;
		return false;
	}

	m_device_delay_counter = 0;

	m_ui.test.SetCurSel(0);
	SendTest(ui::test_off);

	ReportStatus(IDS_DEVICE_OPENED);

	StartStream();
	return true;
}

void CZET030Dlg::CloseDevice()
{
	if (m_device)
	{
		m_ui.test.SetCurSel(0);
		SendTest(ui::test_off);

		StopStream();

		zet030_close(m_device);
		m_device = NULL;
		m_device_ready = false;
		m_device_name.Empty();

		ReportStatus(IDS_DEVICE_CLOSED);
		UpdateControls();
	}
}

void CZET030Dlg::WatchDevice()
{
	if (m_device_path_selected.empty())
	{
		CloseDevice();
		return;
	}

	if (m_device)
	{
		int r = zet030_check_open(m_device);
		if (r == 0)
		{
			if (!m_device_ready)
			{
				m_device_ready = true;
				InitDevice();
			}
		}
		else
		{
			if (!m_device_ready && r == ZET030_ERROR_BUSY)
				return;

			if (!m_device_ready)
				ReportStatus(IDS_DEVICE_OPEN_FAILED);
			CloseDevice();
		}
		return;
	}

	if (!m_device_path_selected.empty())
		OpenDevice();
}

bool CZET030Dlg::OpenServer()
{
#ifdef _WIN64
#pragma message("WARNING! SRV is 32-bit only")
	return false;
#endif

	if (m_srv)
		return true;

	m_srv = std::make_unique<CZetServerCtrl>();
	m_srv->Create(L"", WS_CHILD, CRect(0, 0, 0, 0), this, IDC_SRV);

	if (m_srv->ConnectVrtCh(channel_quantity) < 0)
	{
		m_srv->DestroyWindow();
		m_srv.reset();
		return false;
	}

	UpdateServerChannels();
	return true;
}

void CZET030Dlg::CloseServer()
{
	if (m_srv)
	{
		m_srv->DisconnectVrtCh();
		m_srv->DestroyWindow();
		m_srv.reset();
	}
}

void CZET030Dlg::WatchServer()
{
	if (m_srv)
	{
		// check whether device or server lost
		if (!m_device_ready || m_srv->OrderConnected() < 0)
			CloseServer();
		return;
	}

	if (m_device_ready)
	{
		// check whether server is running
		HANDLE hmap = OpenFileMapping(FILE_MAP_READ, FALSE, L"MySharedData");
		if (hmap != NULL)
		{
			CloseHandle(hmap);
			OpenServer();
		}
	}
}

void CZET030Dlg::UpdateServerChannels(void)
{
	if (!m_srv)
		return;

	auto update = [&](long ch, bool enabled, const CString &group, const CString &name, const CString &unit, float freq, float max_level, float min_level, float reference)
	{
		BSTR sysName = name.AllocSysString();
		BSTR sysUnit = unit.AllocSysString();
		m_srv->VrtChnInfo(ch, &sysName, &sysUnit, freq, max_level, min_level, reference);
		SysFreeString(sysName);
		SysFreeString(sysUnit);

		m_srv->SetGroupName(ch, group);

		m_srv->SetStatus(ch, enabled ? 2 /* virtual */ : 9 /* inactive virtual */);
		if (enabled)
		{
			float dummy = 0;
			m_srv->PutData(ch, 0, 1, &dummy);
		}
	};

	for (long ch = 0; ch < 4; ch++)
	{
		CString s;
		s.Format(L"%s #%d", m_device_name.GetString(), ch + 1);
		update(ch, m_cfg.channel[ch].enabled, m_device_name, s, m_cfg.channel[ch].unit.c_str(), m_state.freq, m_state.channel[ch].max_level, 1e-3f, 1e-3f);
	}

	m_state.time = std::floor(m_srv->WorkingTime());
	m_state.counter64 = 0;
}
