
// ZET030Dlg.h : header file
//

#pragma once

// CZET030Dlg dialog
class CZET030Dlg : public CDialogEx
{
public:
	CZET030Dlg(CWnd* pParent = nullptr);
	virtual ~CZET030Dlg();

// Dialog Data
#ifdef AFX_DESIGN_TIME
	enum { IDD = IDD_ZET030_DIALOG };
#endif

protected:
	static const wchar_t *device_name;
	static const int device_type = 30;
	static const size_t channel_quantity = 4;
	static const char *config_path;
	static const char *eth_addr;
	static const std::vector<uint32_t> freq_table;

	// UI controls
	struct ui
	{
		enum test
		{
			test_off,
			test_sqr,
			test_short,
			test_neg,
			test_sqr_short,
		};

		struct channel
		{
			CStatic icon;
			CStatic name;
			CStatic level;
			CButton icp;
			CButton amp;
			CEdit sense;
			CEdit unit;
		};

		struct ethernet
		{
			CStatic label;
			CComboBox method;
			CEdit addr;
			CButton ftp;
		};

		CComboBox device;
		CButton connect;
		CStatic status;

		std::array<channel, channel_quantity> channel;

		CStatic freq_label;
		CComboBox freq;

		CStatic test_label;
		CComboBox test;

		ethernet eth;

		CStatic rec_mode_label;
		CComboBox rec_mode;

		CStatic rec_minutes_label;
		CEdit rec_minutes;

		CButton apply;
		CButton revert;
	};

	// device configuration
	struct config
	{
		struct channel
		{
			float lsb;
			bool enabled;
			bool icp;
			float amp_ratio;
			std::wstring sense;
			std::wstring unit;

			channel()
				: lsb(1000.0f)
				, enabled(true)
				, icp(false)
				, amp_ratio(1.0f)
				, sense(L"0.001")
				, unit(L"mV")
			{}

			bool operator ==(const struct channel &other) const
			{
				return enabled == other.enabled &&
					icp == other.icp &&
					amp_ratio == other.amp_ratio &&
					sense == other.sense &&
					unit == other.unit;
			}
		};

		struct ethernet
		{
			bool dhcp;
			std::string addr;
			bool ftp;

			ethernet()
				: dhcp(false)
				, ftp(false)
			{}

			bool operator ==(const struct ethernet &other) const
			{
				return dhcp == other.dhcp &&
					addr == other.addr &&
					ftp == other.ftp;
			}

		};

		std::array<channel, channel_quantity> channel;
		uint32_t freq;

		std::optional<ethernet> eth;

		std::optional<bool> rec_button;
		uint32_t rec_minutes;

		config()
			: freq(0)
			, rec_minutes(0)
		{
		}

		bool operator ==(const struct config &other) const
		{
			return channel == other.channel && 
				freq == other.freq &&
				eth == other.eth &&
				rec_button == other.rec_button &&
				rec_minutes == other.rec_minutes;
		}
	};

	// current state (device -> ZETServer)
	struct state
	{
		struct channel
		{
			bool enabled;
			CString name;
			CString unit;
			float int_to_float;
			float max_level;
			double mean_acc;
			double rms_acc;
			size_t stat_cnt;

			channel()
				: enabled(true)
				, int_to_float(0)
				, max_level(0)
				, mean_acc(0)
				, rms_acc(0)
				, stat_cnt(0)
			{}
		};

		bool started;
		float freq;
		std::array<channel, channel_quantity> channel;
		size_t channel_count;
		uint64_t counter64;
		double time;
		uint32_t counter;

		std::mutex mutex;
		std::vector<int32_t> fifo;

		state()
			: started(false)
			, freq(0)
			, channel_count(channel_quantity)
			, time(0)
			, counter(0)
		{}
	};

protected:
	HICON m_icon;
	HICON m_icon_input_off;
	HICON m_icon_input_on;
	HICON m_icon_input_icp;

	struct ui m_ui;

	std::list<CString> m_device_list;
	std::string m_device_path_selected;
	struct zet030_device *m_device;
	CString m_device_name;
	std::atomic_bool m_device_ready;
	uint32_t m_device_delay_counter;

	// config
	struct config m_cfg;
	struct config m_cfg_temp;
	pugi::xml_document m_xml;

	// ZETServer
	std::unique_ptr<CZetServerCtrl> m_srv;

	struct state m_state;

protected:
	DECLARE_MESSAGE_MAP()

	virtual void DoDataExchange(CDataExchange *pDX);
	virtual BOOL OnInitDialog();
	virtual BOOL DestroyWindow();

	afx_msg HCURSOR OnQueryDragIcon();
	afx_msg void OnPaint();
	afx_msg void OnOK();
	afx_msg void OnCancel();
	afx_msg void OnClose();
	afx_msg void OnTimer(UINT_PTR nIDEvent);
	afx_msg void OnCbnSelendokComboDevice();
	afx_msg void OnCbnEditupdateComboDevice();
	afx_msg void OnBnClickedButtonConnect();
	afx_msg void OnStnClickedStaticIcon1();
	afx_msg void OnStnClickedStaticIcon2();
	afx_msg void OnStnClickedStaticIcon3();
	afx_msg void OnStnClickedStaticIcon4();
	afx_msg void OnBnClickedCheckIcp();
	afx_msg void OnBnClickedCheckAmp();
	afx_msg void OnEnChangeEditSense();
	afx_msg void OnEnChangeEditUnit();
	afx_msg void OnCbnSelchangeComboFreq();
	afx_msg void OnCbnSelchangeComboTest();
	afx_msg void OnCbnSelchangeComboEthMethod();
	afx_msg void OnEnChangeEditEthAddr();
	afx_msg void OnBnClickedCheckEthFtp();
	afx_msg void OnCbnSelchangeComboRecMode();
	afx_msg void OnEnChangeEditRecMinutes();
	afx_msg void OnBnClickedButtonApply();
	afx_msg void OnBnClickedButtonRevert();

protected:
	void ReportStatus(const CString &text);
	void ReportStatus(UINT ids);
	void UpdateControls(bool skip_set_value = false);
	void ToggleChannelControl(size_t ch);

	void RefreshDeviceList();
	std::string GetDevicePath();
	bool OpenDevice();
	bool InitDevice();
	void CloseDevice();
	void WatchDevice();

	bool LoadConfig();
	bool SaveConfig();

	bool SendTest(enum ui::test test);

	bool StartStream();
	bool StopStream();
	void ReceiveStream(const int32_t *data, int count);
	void WatchStream();

	bool OpenServer();
	void CloseServer();
	void WatchServer();
	void UpdateServerChannels(void);
};
